#define _GNU_SOURCE

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <pthread.h>
#include <unistd.h>

struct rte_thread_ctrl_params {
	void *(*start_routine)(void *);
	void *arg;
	pthread_barrier_t configured;
};

int rte_thread_setname(pthread_t id, const char *name)
{
	int ret = -1;

	(void)id;
	(void)name;
	ret = pthread_setname_np(id, name);
	return ret;
}

static void *rte_thread_init(void *arg)
{
	struct rte_thread_ctrl_params *params = arg;
	void *(*start_routine)(void *) = params->start_routine;
	void *routine_arg = params->arg;
	int ret;

	printf("rte_thread_init\n");

	printf("child wait\n");
	ret = pthread_barrier_wait(&params->configured);
	if (ret == PTHREAD_BARRIER_SERIAL_THREAD) {
		printf("child destroy\n");
		pthread_barrier_destroy(&params->configured);
		free(params);
	}

	return start_routine(routine_arg);
}

int
rte_ctrl_thread_create(pthread_t *thread, const char *name,
		const pthread_attr_t *attr,
		void *(*start_routine)(void *), void *arg)
{
	struct rte_thread_ctrl_params *params;
	unsigned int lcore_id;
	cpu_set_t cpuset;
	int ret;

	printf("rte_ctrl_thread_create\n");

	params = malloc(sizeof(*params));
	if (!params)
		return -1;

	params->start_routine = start_routine;
	params->arg = arg;

	pthread_barrier_init(&params->configured, NULL, 2);

	ret = pthread_create(thread, attr, rte_thread_init, (void *)params);
	if (ret != 0)
		return ret;

	if (name != NULL) {
		ret = rte_thread_setname(*thread, name);
		ret = -1; /* FORCE FAILURE */
		if (ret < 0)
			goto fail;
	}

	CPU_ZERO(&cpuset);
	for (lcore_id = 0; lcore_id < 4; lcore_id++) {
		if (lcore_id >= 2)
			CPU_SET(lcore_id, &cpuset);
	}

	ret = pthread_setaffinity_np(*thread, sizeof(cpuset), &cpuset);
	if (ret < 0)
		goto fail;

	ret = pthread_barrier_wait(&params->configured);
	if (ret == PTHREAD_BARRIER_SERIAL_THREAD) {
		pthread_barrier_destroy(&params->configured);
		free(params);
	}

	return 0;

fail:
#ifdef FIX
	printf("parent wait\n");
	ret = pthread_barrier_wait(&params->configured);
	if (ret == PTHREAD_BARRIER_SERIAL_THREAD) {
		printf("parent destroy\n");
		pthread_barrier_destroy(&params->configured);
		free(params);
	}
#else
	pthread_barrier_destroy(&params->configured);
	free(params);
#endif
	printf("parent thread cancel\n");
	pthread_cancel(*thread);
	pthread_join(*thread, NULL);
	return ret;
}

static void *start(void *arg)
{
	(void)arg;
	printf("child waiting 8s\n");
	sleep(8);
	printf("child (not canceled)\n");
	return NULL;
}

int main(void)
{
	pthread_t thread;
	int ret;

	printf("main\n");
	ret = rte_ctrl_thread_create(&thread, "test", NULL, start, NULL);
	printf("parent thread create ret=%d\n", ret);

	printf("parent waiting 10s\n");
	sleep(10);

	return 0;
}
